/*
 *  Copyright (C) 2021-2024 Texas Instruments Incorporated
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *    Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the
 *    distribution.
 *
 *    Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 *  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 *  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <kernel/dpl/CacheP.h>
#include <kernel/dpl/DebugP.h>
#include <kernel/dpl/SemaphoreP.h>
#include "ti_drivers_config.h"
#include "ti_drivers_open_close.h"
#include "ti_board_open_close.h"

/*
 * This example performs UDMA block copy transfer using Type 15 Transfer Record (TR15)
 * using Transfer Record Packet Descriptor (TRPD) in interrupt mode.
 *
 * The application opens and configures a BCDMA channel using SysConfig.
 * It also configures the interrupt mode of operation through the SysConfig
 * which ensures that all required interrupt configuration are done.
 * The callback function App_udmaEventCb is registered via SysConfig.
 *
 * Then the application prepares a TRPD for a 1D transfer from source to
 * destination buffer, submits the request to DMA, waits for the DMA to complete
 * by waiting on a semaphore which is posted in the callback function.
 *
 * Once the transfer it completes, it does cache operation for data coherency
 * and compares the source and destination buffers for any data mismatch.
 *
 */


extern void ipc_rpmsg_echo_main_core_start(void);
extern void ipc_rpmsg_echo_main_core_receive(void);
void ipc_rpmsg_echo_main_core_send(uint64_t GTCReadTime, uint32_t channel);



/* This example is applicable for even (1.2,4,8,16) count of TRPDs only */
#define NUM_CHANNEL CONFIG_UDMA0_NUM_BLKCOPY_CH //it should be same as CONFIG_UDMA0_NUM_BLKCOPY_CH

#define channel_1       0
#define channel_2       1
#define channel_3       2
#define channel_4       3

/* Number of bytes to do memcpy */
#define UDMA_TEST_NUM_BYTES             (4 * 1024U)
/* UDMA TR packet descriptor memory size - with one TR */
#define UDMA_TEST_TRPD_SIZE             (UDMA_GET_TRPD_TR15_SIZE(1U))

/* UDMA TRPD Memory */
uint8_t gUdmaTestTrpdMem[NUM_CHANNEL][UDMA_TEST_TRPD_SIZE] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT)));

extern volatile uint32_t  Read[NUM_CHANNEL];

#define UART_PRINT 0x00
#define TEST 0x00


#if NUM_CHANNEL > 4
    #define NUM_CHANNEL_8
    #define channel_5       4
    #define channel_6       5
    #define channel_7       6
    #define channel_8       7

#endif

uint64_t msgValue = 0x5A5A;


Udma_DrvHandle      drvHandle = &gUdmaDrvObj[0];
Udma_EventPrms      eventPrms;

Udma_EventObject    gUdmaCqEventObj[NUM_CHANNEL];

Udma_ChHandle       chHandle[NUM_CHANNEL];
 uint8_t     *trpdMem[NUM_CHANNEL];
 uint64_t    trpdMemPhy[NUM_CHANNEL];

static volatile uint32_t completed_channel_transfer[NUM_CHANNEL] ={0x00,0x00,0x00,0x00}; //completed transfer on channel 1-4

#if(UART_PRINT == 0x01)
static uint32_t Compare_Buf_Status[NUM_CHANNEL] ={0x00,0x00,0x00,0x00}; //completed transfer on channel 1-4
#endif

/* NUmber of Iterations */
#define NUM_ITER  1 // max value 10 (BUF_SIZE/MAX_PACKET_SIZE)

/* Application Buffers */
uint8_t gUdmaTestSrcBuf0_0[UDMA_ALIGN_SIZE(UDMA_TEST_NUM_BYTES)] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT)));
uint8_t gUdmaTestSrcBuf0_1[UDMA_ALIGN_SIZE(UDMA_TEST_NUM_BYTES)] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT)));
uint8_t gUdmaTestSrcBuf1_0[UDMA_ALIGN_SIZE(UDMA_TEST_NUM_BYTES)] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT)));
uint8_t gUdmaTestSrcBuf1_1[UDMA_ALIGN_SIZE(UDMA_TEST_NUM_BYTES)] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT)));


uint8_t gUdmaTestDestBuf_0_0[UDMA_ALIGN_SIZE(UDMA_TEST_NUM_BYTES)] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss.R5F0_0_TCMA_filebuf")));
uint8_t gUdmaTestDestBuf_0_1[UDMA_ALIGN_SIZE(UDMA_TEST_NUM_BYTES)] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss.R5F0_1_TCMB_filebuf")));
uint8_t gUdmaTestDestBuf_1_0[UDMA_ALIGN_SIZE(UDMA_TEST_NUM_BYTES)] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss.R5F1_0_TCMA_filebuf")));
uint8_t gUdmaTestDestBuf_1_1[UDMA_ALIGN_SIZE(UDMA_TEST_NUM_BYTES)] __attribute__((aligned(UDMA_CACHELINE_ALIGNMENT), section(".bss.R5F1_1_TCMB_filebuf")));


/* Semaphore to indicate transfer completion */
static SemaphoreP_Object gUdmaTestDoneSem;

void App_udmaEventCb(Udma_EventHandle eventHandle, uint32_t eventType, void *appData);
static void App_udmaTrpdInit(Udma_ChHandle chHandle,
                             uint8_t *trpdMem,
                             const void *destBuf,
                             const void *srcBuf,
                             uint32_t length);
static void App_udmaInitBuf(uint8_t *srcBuf, uint8_t *destBuf, uint32_t length, uint32_t channel,uint32_t flag);
static void App_udmaCompareBuf(uint8_t *srcBuf, uint8_t *destBuf, uint32_t length, uint32_t channel);

void app_check_tr_response(uint32_t channel);
volatile uint64_t        GTCReadTime[NUM_CHANNEL] = {0U};

uint8_t        *srcBuf[NUM_CHANNEL];
uint8_t        *destBuf[NUM_CHANNEL];

volatile uint8_t        DMA_Completion_Event_flag = 0xFF;
uint32_t       DMA_Startcount = 0x00;


void *udma_memcpy_interrupt_main(void *args)
{
    int32_t         retVal = UDMA_SOK;

    uint32_t        length = UDMA_TEST_NUM_BYTES;
    //uint8_t        *trpdMem[NUM_CHANNEL];
    uint32_t channel = 0U;

    /* Submit TRPD to channel */
#if(UART_PRINT == 0x01)
    uint32_t Time_MSB = 0x00;
    uint32_t Time_LSB = 0x00;
#endif
    uint32_t gBaseAddr = 0x00;


    //Map R5F TCM memory to destination buffer array
    destBuf[0] = &gUdmaTestDestBuf_0_0[0U];
    destBuf[1] = &gUdmaTestDestBuf_0_1[0U];
    destBuf[2] = &gUdmaTestDestBuf_1_0[0U];
    destBuf[3] = &gUdmaTestDestBuf_1_1[0U];

    srcBuf[0] = &gUdmaTestSrcBuf0_0[0U];
    srcBuf[1] = &gUdmaTestSrcBuf0_1[0U];
    srcBuf[2] = &gUdmaTestSrcBuf1_0[0U];
    srcBuf[3] = &gUdmaTestSrcBuf1_1[0U];


    #if defined (SOC_AM65X)
    uint32_t        instId;
    #endif


    /* Open drivers to open the UART driver for console */
    Drivers_open();
    Board_driversOpen();


    /* Get address after translation translate */
     gBaseAddr = (uint32_t) AddrTranslateP_getLocalAddr(CONFIG_GPIO0_BASE_ADDR);
     GPIO_setDirMode(gBaseAddr, CONFIG_GPIO0_PIN, CONFIG_GPIO0_DIR);
     GPIO_pinWriteLow(gBaseAddr, CONFIG_GPIO0_PIN);


    for(channel =0; channel < NUM_CHANNEL; channel++)
    {

        chHandle[channel] = gConfigUdma0BlkCopyChHandle[channel];
        trpdMem[channel] = gUdmaTestTrpdMem[channel];
        trpdMemPhy[channel] = (uint64_t) Udma_defaultVirtToPhyFxn(trpdMem[channel], 0U, NULL);

        DebugP_log("[UDMA] Memcpy application started on A53 for channel = %d...\r\n", channel);

        /* Channel enable */
        retVal = Udma_chEnable(chHandle[channel]);
        DebugP_assert(UDMA_SOK == retVal);


        /* Init buffers and TR packet descriptor */
        App_udmaInitBuf(srcBuf[channel], destBuf[channel], length, channel, 1U);
        App_udmaTrpdInit(chHandle[channel], trpdMem[channel], destBuf[channel], srcBuf[channel], length);

    }

    #if(TEST == 0x00)
    ipc_rpmsg_echo_main_core_start();
    #endif

    TimerP_start( gTimerBaseAddr[CONFIG_TIMER0] );

    while(1)
    {
        
      if((completed_channel_transfer[channel_1] == 1) && (completed_channel_transfer[channel_2] == 1) && (completed_channel_transfer[channel_3] == 1) && (completed_channel_transfer[channel_4] == 1))
      {

#if(TEST == 0x00)

        if( (Read[0U] == 0x01) && (Read[1U] == 0x02) && (Read[2U] == 0x03) && (Read[3U] == 0x04) )
#endif
        {

                for(channel =0; channel < NUM_CHANNEL; channel++)
                {
                    Read[channel] = 0x00;
                    completed_channel_transfer[channel] = 0x00;

                    app_check_tr_response(channel);

                 #if(UART_PRINT == 0x01)

                   if ( Compare_Buf_Status[channel] == 0x00)
                   {
                       DebugP_log("Source and Destination buffers are the same  for Channel = %d ...\r\n", channel);
                   }
                   else
                   {
                       DebugP_log("Source and Destination buffers are mismatch for Channel = %d ...\r\n", channel);
                   }
                 #endif

                   /* Init buffers and TR packet descriptor */
                   App_udmaInitBuf(srcBuf[channel], destBuf[channel], length, channel, 1U);
                }

            #if(UART_PRINT == 0x01)
            DebugP_log("DMA start count on A53  %x  ...\r\n", DMA_Startcount);
            for(channel =0; channel < NUM_CHANNEL; channel++)
            {
                Time_MSB = GTCReadTime[channel] >> 32U;
                Time_LSB = GTCReadTime[channel] & 0x00000000FFFFFFFF;
                DebugP_log("DMA start event for Channel = %d && Time MSB = %x && Time LSB = %x ...\r\n", channel, Time_MSB,Time_LSB);
                DebugP_log("DMA start event for channel in usec  = %f    ...\r\n", ((float)(GTCReadTime[channel]) ) / 1U );
            }
            #endif

            DMA_Completion_Event_flag = 0xFF;

            #if(TEST == 0x01)
            *( (volatile uint32_t *)0x0060106C ) = 4U;
            #endif

            //ClockP_sleep(1);

        }


      }

    }

    /* Channel disable */
    for(channel =0; channel < NUM_CHANNEL; channel++)
    {
    /* Channel disable */
        retVal = Udma_chDisable(chHandle[channel], UDMA_DEFAULT_CH_DISABLE_TIMEOUT);
        DebugP_assert(UDMA_SOK == retVal);
    }

    SemaphoreP_destruct(&gUdmaTestDoneSem);
    DebugP_log("All tests have passed!!\r\n");
    Board_driversClose();
    Drivers_close();
    #if defined (SOC_AM65X)
    for(instId = 0U; instId < CONFIG_UDMA_NUM_INSTANCES; instId++)
    {
       retVal = Udma_deinit(&gUdmaDrvObj[instId]);
       DebugP_assert(UDMA_SOK == retVal);
    }
    #endif

    return NULL;
}

void App_udmaEventCb0(Udma_EventHandle eventHandle, uint32_t eventType, void *appData)
{
    if(UDMA_EVENT_TYPE_DMA_COMPLETION == eventType)
    {

            //app_check_tr_response(channel_1);
            completed_channel_transfer[0] = 1;
#if(TEST == 0x00)
            ipc_rpmsg_echo_main_core_send(msgValue, channel_1);
#endif
    }
}

void App_udmaEventCb1(Udma_EventHandle eventHandle, uint32_t eventType, void *appData)
{
    if(UDMA_EVENT_TYPE_DMA_COMPLETION == eventType)
    {

            //app_check_tr_response(channel_2);
            completed_channel_transfer[1] = 1;
#if(TEST == 0x00)

            ipc_rpmsg_echo_main_core_send(msgValue, channel_2);
#endif
    }
}

void App_udmaEventCb2(Udma_EventHandle eventHandle, uint32_t eventType, void *appData)
{
    if(UDMA_EVENT_TYPE_DMA_COMPLETION == eventType)
    {

            //app_check_tr_response(channel_3);
            completed_channel_transfer[2] = 1;
#if(TEST == 0x00)

            ipc_rpmsg_echo_main_core_send(msgValue, channel_3);
#endif
    }
}

void App_udmaEventCb3(Udma_EventHandle eventHandle, uint32_t eventType, void *appData)
{
    if(UDMA_EVENT_TYPE_DMA_COMPLETION == eventType)
    {

            //app_check_tr_response(channel_4);
            completed_channel_transfer[3] = 1;
#if(TEST == 0x00)

            ipc_rpmsg_echo_main_core_send(msgValue, channel_4);
#endif
    }

}


void app_check_tr_response(uint32_t channel)
{
    uint64_t        pDesc;
    int32_t status = SystemP_SUCCESS;


    status = Udma_ringDequeueRaw(Udma_chGetCqRingHandle(chHandle[channel]), &pDesc);

      if(UDMA_SOK == status)
      {
          /* Check TR response status */
          CacheP_inv(trpdMem[channel], UDMA_TEST_TRPD_SIZE, CacheP_TYPE_ALLD);
          status = UdmaUtils_getTrpdTr15Response(trpdMem[channel], 1U, 0U);

          if(status != CSL_UDMAP_TR_RESPONSE_STATUS_COMPLETE)
          {
#if(UART_PRINT == 0x01)

             DebugP_log("TR Response failed for transfer : channel = %u\r\n", channel);
#endif
             DebugP_assert(FALSE);
          }
          else
          {
             /* Invalidate destination buffer */
            CacheP_inv(destBuf[channel], UDMA_TEST_NUM_BYTES, CacheP_TYPE_ALLD);
            App_udmaCompareBuf(srcBuf[channel], destBuf[channel], UDMA_TEST_NUM_BYTES,channel);
#if(UART_PRINT == 0x01)

            DebugP_log("TR Response Success for transfer : channel = %u\r\n", channel);
#endif
          }
      }

}

static void App_udmaTrpdInit(Udma_ChHandle chHandle,
                             uint8_t *trpdMem,
                             const void *destBuf,
                             const void *srcBuf,
                             uint32_t length)
{
    CSL_UdmapTR15  *pTr;
    uint32_t        cqRingNum = Udma_chGetCqRingNum(chHandle);

    /* Make TRPD with TR15 TR type */
    UdmaUtils_makeTrpdTr15(trpdMem, 1U, cqRingNum);

    /* Setup TR */
    pTr = UdmaUtils_getTrpdTr15Pointer(trpdMem, 0U);
    pTr->flags    = CSL_FMK(UDMAP_TR_FLAGS_TYPE, CSL_UDMAP_TR_FLAGS_TYPE_4D_BLOCK_MOVE_REPACKING_INDIRECTION);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_STATIC, 0U);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_EOL, CSL_UDMAP_TR_FLAGS_EOL_MATCH_SOL_EOL);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_EVENT_SIZE, CSL_UDMAP_TR_FLAGS_EVENT_SIZE_COMPLETION);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_TRIGGER0, CSL_UDMAP_TR_FLAGS_TRIGGER_NONE);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_TRIGGER0_TYPE, CSL_UDMAP_TR_FLAGS_TRIGGER_TYPE_ALL);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_TRIGGER1, CSL_UDMAP_TR_FLAGS_TRIGGER_NONE);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_TRIGGER1_TYPE, CSL_UDMAP_TR_FLAGS_TRIGGER_TYPE_ALL);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_CMD_ID, 0x25U);  /* This will come back in TR response */
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_SA_INDIRECT, 0U);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_DA_INDIRECT, 0U);
    pTr->flags   |= CSL_FMK(UDMAP_TR_FLAGS_EOP, 1U);
    pTr->icnt0    = length;
    pTr->icnt1    = 1U;
    pTr->icnt2    = 1U;
    pTr->icnt3    = 1U;
    pTr->dim1     = pTr->icnt0;
    pTr->dim2     = (pTr->icnt0 * pTr->icnt1);
    pTr->dim3     = (pTr->icnt0 * pTr->icnt1 * pTr->icnt2);
    pTr->addr     = (uint64_t) Udma_defaultVirtToPhyFxn(srcBuf, 0U, NULL);
    pTr->fmtflags = 0x00000000U;    /* Linear addressing, 1 byte per elem */
    pTr->dicnt0   = length;
    pTr->dicnt1   = 1U;
    pTr->dicnt2   = 1U;
    pTr->dicnt3   = 1U;
    pTr->ddim1    = pTr->dicnt0;
    pTr->ddim2    = (pTr->dicnt0 * pTr->dicnt1);
    pTr->ddim3    = (pTr->dicnt0 * pTr->dicnt1 * pTr->dicnt2);
    pTr->daddr    = (uint64_t) Udma_defaultVirtToPhyFxn(destBuf, 0U, NULL);

    /* Perform cache writeback */
    CacheP_wb(trpdMem, UDMA_TEST_TRPD_SIZE, CacheP_TYPE_ALLD);

    return;
}

static void App_udmaInitBuf(uint8_t *srcBuf, uint8_t *destBuf, uint32_t length, uint32_t channel,uint32_t flag)
{
    uint32_t        i;

    for(i = 0U; i < length; i++)
    {
        srcBuf[i] = i + channel;
        destBuf[i] = 0xA5U + channel;
    }

    if(flag)
    {
        /* Writeback source and destination buffer */
        CacheP_wb(srcBuf, length, CacheP_TYPE_ALLD);
        CacheP_wb(destBuf, length, CacheP_TYPE_ALLD);
    }


    return;
}

static void App_udmaCompareBuf(uint8_t *srcBuf, uint8_t *destBuf, uint32_t length, uint32_t channel)
{
    uint32_t        i;

#if(UART_PRINT == 0x01)
    Compare_Buf_Status[channel] = 0x00;
#endif

    //CacheP_inv(destBuf, length, CacheP_TYPE_ALLD);
    for(i = 0U; i < length; i++)
    {
        if(srcBuf[i] != destBuf[i])
        {
#if(UART_PRINT == 0x01)

            Compare_Buf_Status[channel] = 0xFF;
#endif
        }
    }

    return;
}


void Timer_ISR(void *args)
{


    uint32_t channel;
    int32_t retVal;

    if(DMA_Completion_Event_flag == 0xFF)
    {
        DMA_Completion_Event_flag = 0x00;

        for(channel =0; channel < NUM_CHANNEL; channel++)
        {
            Read[channel] = 0x00;
            completed_channel_transfer[channel] = 0x00;

            /* Submit TRPD to channel */
            GTCReadTime[channel] = GTC_getCount64();

            *( (volatile uint32_t *)0x00601068 ) = 4U;
            retVal = Udma_ringQueueRaw(Udma_chGetFqRingHandle(chHandle[channel]), trpdMemPhy[channel]);
            DebugP_assert(UDMA_SOK == retVal);
        }

        DMA_Startcount++;

    }



}
